/****************************************************************************************[Solver.C]
MiniSat -- Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#include <omp.h>
#include "Solver.h"

#include <vector>

#ifdef _MSC_VER
 #include "../mtl/Sort.h"
 #include <math.h>
#include <ctime>
#else
 #include "Sort.h"
#endif

#include <cmath>

static double C11 = 1200;
static double C22 = 1200;
//=================================================================================================
// Constructor/Destructor:


Solver::Solver() :

    // Parameters: (formerly in 'SearchParams')
    var_decay(1 / 0.95), clause_decay(1 / 0.999), random_var_freq(0.02)
  , restart_first(100), restart_inc(1.5), learntsize_factor((double)1/(double)3), learntsize_inc(1.1)

    // More parameters:
    //
  , expensive_ccmin  (true)
  , polarity_mode    (polarity_user)
  , verbosity        (0)

    // Statistics: (formerly in 'SolverStats')
    //
  , starts(0), decisions(0), rnd_decisions(0), propagations(0), conflicts(0)
  , clauses_literals(0), learnts_literals(0), max_literals(0), tot_literals(0)

  , ok               (true)
  , cla_inc          (1)
  , var_inc          (1)
  , qhead            (0)
  , simpDB_assigns   (-1)
  , simpDB_props     (0)
  , order_heap       (VarOrderLt(activity))
  , random_seed      (91648253)
  , progress_estimate(0)
  , remove_satisfied (true)
{	//multi-threading
  nbExportedClauses  = nbExportedUnits = 0;nbConf = 0;max_act=0; prior_limit = 5;}


Solver::~Solver()
{
    for (int i = 0; i < learnts.size(); i++) free(learnts[i]);
    for (int i = 0; i < clauses.size(); i++) free(clauses[i]);
}


//=================================================================================================
// Minor methods:


// Creates a new SAT variable in the solver. If 'decision_var' is cleared, variable will not be
// used as a decision variable (NOTE! This has effects on the meaning of a SATISFIABLE result).
//
Var Solver::newVar(bool sign, bool dvar)
{
    int v = nVars();
    watches   .push();          // (list for positive literal)
    watches   .push();          // (list for negative literal)
    reason    .push(NULL);
    assigns   .push(toInt(l_Undef));
    level     .push(-1);
    activity  .push(0);
    seen      .push(0);
    seen2      .push(0);
    seen3      .push(0);
    score      .push(0.0);

    polarity    .push((char)sign);
    decision_var.push((char)dvar);

    insertVarOrder(v);
    return v;
}


bool Solver::addClause(vec<Lit>& ps)
{
    assert(decisionLevel() == 0);

    if (!ok)
        return false;
    else{
        // Check if clause is satisfied and remove false/duplicate literals:
        sort(ps);
        Lit p; int i, j;
        for (i = j = 0, p = lit_Undef; i < ps.size(); i++)
	  if (value(ps[i]) == l_True || ps[i] == ~p)
	    return true;
	  else if (value(ps[i]) != l_False && ps[i] != p)
	    ps[j++] = p = ps[i];
        ps.shrink(i - j);
    }

    if (ps.size() == 0)
        return ok = false;
    else if (ps.size() == 1){
        assert(value(ps[0]) == l_Undef);
        uncheckedEnqueue(ps[0]);
        return ok = (propagate() == NULL);
    }else{
      Clause* c = Clause_new(ps, false, threadId);
        clauses.push(c);
        attachClause(*c);
    }

    return true;
}

void Solver::attachClause(Clause& c) {

  Bclause bcl;
  bcl.a = c[0];
  bcl.cl = &c;
  
  assert(c.size() > 1);
  watches[toInt(~c[0])].push(bcl);
  watches[toInt(~c[1])].push(bcl);
  if (c.learnt()) learnts_literals += c.size();
  else            clauses_literals += c.size(); }


void Solver::detachClause(Clause& c) {


  assert(c.size() > 1);
  
  vec<Bclause> & ts = watches[toInt(~c[0])];
  int j = 0;
  for (; j < ts.size() && ts[j].cl != &c; j++);
  assert(j < ts.size());
  for (; j < ts.size()-1; j++) ts[j] = ts[j+1];
  ts.pop();


  vec<Bclause> & ts2 = watches[toInt(~c[1])];
  j = 0;
  for (; j < ts2.size() && ts2[j].cl != &c; j++);
  assert(j < ts2.size());
  for (; j < ts2.size()-1; j++) ts2[j] = ts2[j+1];
  ts2.pop();
  if (c.learnt()) learnts_literals -= c.size();
  else            clauses_literals -= c.size();
}

void Solver::removeClause(Clause& c) {
    detachClause(c);
    free(&c); }


bool Solver::satisfied(const Clause& c) const {
    for (int i = 0; i < c.size(); i++)
        if (value(c[i]) == l_True)
            return true;
    return false; }


// Revert to the state at given level (keeping all assignment at 'level' but not beyond).
//
void Solver::cancelUntil(int level) {
  if (decisionLevel() > level){
    for (int c = trail.size()-1; c >= trail_lim[level]; c--){
      Var     x  = var(trail[c]);
      assigns[x] = toInt(l_Undef);
      
      if(threadId % 4 == 1 || threadId %4 == 0)
	polarity[x] = sign(trail[c]);
      
      insertVarOrder(x); }
    qhead = trail_lim[level];
    trail.shrink(trail.size() - trail_lim[level]);
    trail_lim.shrink(trail_lim.size() - level);
  } }


//=================================================================================================
// Major methods:


Lit Solver::pickBranchLit(int polarity_mode, double random_var_freq, Cooperation* C)
{
    Var next = var_Undef;
    
    if(threadId >= 4){
      for(int i = 0; i < prior_lits.size();i++){
	Lit q = prior_lits[i];
	if(value(q) == l_Undef)
	  return q;
      }
    }

    // Random decision:
    if (drand(threadId, random_seed) < random_var_freq && !order_heap.empty()){
      next = order_heap[irand(threadId, random_seed,order_heap.size())];
      if (toLbool(assigns[next]) == l_Undef && decision_var[next])
	rnd_decisions++; }
    
    // Activity based decision:
    while (next == var_Undef || toLbool(assigns[next]) != l_Undef || !decision_var[next])
      if (order_heap.empty()){
		next = var_Undef;
		break;
      }else
	next = order_heap.removeMin();
    
    bool sign = false;
    switch (polarity_mode){
		case polarity_true:  sign = false; break;
		case polarity_false: sign = true;  break;
		case polarity_user:  sign = polarity[next]; break;
		case polarity_rnd:   sign = irand(threadId, random_seed, 2); break;
		default: assert(false); 
	}
    
    if((next != var_Undef) && (threadId % 4 == 3))
      polarity[next] = (score[next] > 0);

    return next == var_Undef ? lit_Undef : Lit(next, sign);
}


/*_________________________________________________________________________________________________
|
|  analyze : (confl : Clause*) (out_learnt : vec<Lit>&) (out_btlevel : int&)  ->  [void]
|  
|  Description:
|    Analyze conflict and produce a reason clause.
|  
|    Pre-conditions:
|      * 'out_learnt' is assumed to be cleared.
|      * Current decision level must be greater than root level.
|  
|    Post-conditions:
|      * 'out_learnt[0]' is the asserting literal at level 'out_btlevel'.
|  
|  Effect:
|    Will undo part of the trail, upto but not beyond the assumption of the current decision level.
|________________________________________________________________________________________________@*/

void Solver::Ranalyze(Clause* confl, vec<Lit>& out_learnt, int& out_btlevel)
{
    int pathC = 0;
    Lit p     = lit_Undef;
    // Generate conflict clause:
    //
    out_learnt.push();      // (leave room for the asserting literal)
    int index   = trail.size() - 1;
    out_btlevel = 0;
    
    int kd, ke,kv;
    bool okk;

    do{
        assert(confl != NULL);          // (otherwise should be UIP)
        Clause& c = *confl;

        if (c.learnt())
            claBumpActivity(c);



	okk = false;
	kv = kd = ke = 0;
	
	for (int j = (p == lit_Undef) ? 0 : 1; j < c.size(); j++){
	  Lit q = c[j];
	  
	  if(level[var(q)] >= decisionLevel()) kv++;
	  
	  if(seen[var(q)] && level[var(q)] >= decisionLevel()) kd++;
	  
	  if(seen[var(q)] && level[var(q)] < decisionLevel()) ke++;
	  
	}

	
	if((kd >= pathC) && (pathC > 0) && (ke >= (out_learnt.size()-1)))
	  { okk = true;}
	
		



        for (int j = (p == lit_Undef) ? 0 : 1; j < c.size(); j++){
            Lit q = c[j];

            if (!seen[var(q)] && level[var(q)] > 0){
                varBumpActivity(var(q));
                seen[var(q)] = 1;
             
		if (level[var(q)] >= decisionLevel())
                    pathC++;
                else{
		  out_learnt.push(q);
		  if (level[var(q)] > out_btlevel)
		    out_btlevel = level[var(q)];
                }
            }
        }



	if((okk == true) && (kv > 1))
          {
            int dd = 0;
            for(int m = 2; m < c.size() ;m++)
              if(level[var(c[m])] >= decisionLevel()){
                dd = m;
                break;
              }
            detachClause(c);
            c[0] = c[dd];
	    c[dd] = c[c.size()-1];
	    c.pop();
            attachClause(c);
          }

	/*
	if((okk == true) && (kv == 1) && (c.size() > 2))
          {
	    BTS++;
	    detachClause(c);
            c[0] = c[c.size()-1];
	    c.pop();
            attachClause(c);
	  }
	*/




        // Select next clause to look at:
        while (!seen[var(trail[index--])]);
        p     = trail[index+1];
	confl = reason[var(p)];
        seen[var(p)] = 0;
        pathC--;
	
    }while (pathC > 0);
    out_learnt[0] = ~p;
    
    
    
    // Simplify conflict clause:
    //
    int i, j;
    if (expensive_ccmin){
        uint32_t abstract_level = 0;
        for (i = 1; i < out_learnt.size(); i++)
            abstract_level |= abstractLevel(var(out_learnt[i])); // (maintain an abstraction of levels involved in conflict)

        out_learnt.copyTo(analyze_toclear);
        for (i = j = 1; i < out_learnt.size(); i++)
            if (reason[var(out_learnt[i])] == NULL || !litRedundant(out_learnt[i], abstract_level))
                out_learnt[j++] = out_learnt[i];
    }else{
        out_learnt.copyTo(analyze_toclear);
        for (i = j = 1; i < out_learnt.size(); i++){
            Clause& c = *reason[var(out_learnt[i])];
            for (int k = 1; k < c.size(); k++)
                if (!seen[var(c[k])] && level[var(c[k])] > 0){
                    out_learnt[j++] = out_learnt[i];
                    break; }
        }
    }
    max_literals += out_learnt.size();
    out_learnt.shrink(i - j);
    tot_literals += out_learnt.size();

    // Find correct backtrack level:
    //
    if (out_learnt.size() == 1)
        out_btlevel = 0;
    else{
        int max_i = 1;
        for (int i = 2; i < out_learnt.size(); i++)
            if (level[var(out_learnt[i])] > level[var(out_learnt[max_i])])
                max_i = i;
        Lit p             = out_learnt[max_i];
        out_learnt[max_i] = out_learnt[1];
        out_learnt[1]     = p;
        out_btlevel       = level[var(p)];
    }

    if(threadId % 4 == 3)
      for(int j = 0; j < out_learnt.size();j++){
	if(sign(out_learnt[j]))
	  score[var(out_learnt[j])] += 0.00001;
	else
	  score[var(out_learnt[j])] -= 0.00001;
      }

    for (int j = 0; j < analyze_toclear.size(); j++) seen[var(analyze_toclear[j])] = 0;    // ('seen[]' is now cleared)
}



void Solver::analyze(Clause* confl, vec<Lit>& out_learnt, int& out_btlevel)
{
    int pathC = 0;
    Lit p     = lit_Undef;
     bool ok6 = true;
     if(threadId %4 == 3)
       Graph_Lit.clear();
    // Generate conflict clause:
    //
    out_learnt.push();      // (leave room for the asserting literal)
    int index   = trail.size() - 1;
    out_btlevel = 0;

    do{
        assert(confl != NULL);          // (otherwise should be UIP)
        Clause& c = *confl;

        if (c.learnt())
            claBumpActivity(c);

        for (int j = (p == lit_Undef) ? 0 : 1; j < c.size(); j++){
            Lit q = c[j];

            if (!seen[var(q)] && level[var(q)] > 0){
                varBumpActivity(var(q));
                seen[var(q)] = 1;
             
		if (level[var(q)] >= decisionLevel())
                    pathC++;
                else{
		  out_learnt.push(q);
		  if (level[var(q)] > out_btlevel)
		    out_btlevel = level[var(q)];
                }
            }
        }

        // Select next clause to look at:
        while (!seen[var(trail[index--])]);
        p     = trail[index+1];
	if(threadId %4 == 3)
	  Graph_Lit.push(p);
	if(ok6){
	  conflict_lit = p;
	  ok6 = false;
	}
	confl = reason[var(p)];
        seen[var(p)] = 0;
        pathC--;

    }while (pathC > 0);
    out_learnt[0] = ~p;
    
    
    
    // Simplify conflict clause:
    //
    int i, j;
    if (expensive_ccmin){
        uint32_t abstract_level = 0;
        for (i = 1; i < out_learnt.size(); i++)
            abstract_level |= abstractLevel(var(out_learnt[i])); // (maintain an abstraction of levels involved in conflict)

        out_learnt.copyTo(analyze_toclear);
        for (i = j = 1; i < out_learnt.size(); i++)
            if (reason[var(out_learnt[i])] == NULL || !litRedundant(out_learnt[i], abstract_level))
                out_learnt[j++] = out_learnt[i];
    }else{
        out_learnt.copyTo(analyze_toclear);
        for (i = j = 1; i < out_learnt.size(); i++){
            Clause& c = *reason[var(out_learnt[i])];
            for (int k = 1; k < c.size(); k++)
                if (!seen[var(c[k])] && level[var(c[k])] > 0){
                    out_learnt[j++] = out_learnt[i];
                    break; }
        }
    }
    max_literals += out_learnt.size();
    out_learnt.shrink(i - j);
    tot_literals += out_learnt.size();

    // Find correct backtrack level:
    //
    if (out_learnt.size() == 1)
        out_btlevel = 0;
    else{
        int max_i = 1;
        for (int i = 2; i < out_learnt.size(); i++)
            if (level[var(out_learnt[i])] > level[var(out_learnt[max_i])])
                max_i = i;
        Lit p             = out_learnt[max_i];
        out_learnt[max_i] = out_learnt[1];
        out_learnt[1]     = p;
        out_btlevel       = level[var(p)];
    }

    if(threadId % 4 == 3)
      for(int j = 0; j < out_learnt.size();j++){
	if(sign(out_learnt[j]))
	  score[var(out_learnt[j])] += 0.00001;
	else
	  score[var(out_learnt[j])] -= 0.00001;
      }

    for (int j = 0; j < analyze_toclear.size(); j++) seen[var(analyze_toclear[j])] = 0;    // ('seen[]' is now cleared)
}


// Check if 'p' can be removed. 'abstract_levels' is used to abort early if the algorithm is
// visiting literals at levels that cannot be removed later.
bool Solver::litRedundant(Lit p, uint32_t abstract_levels)
{
    analyze_stack.clear(); analyze_stack.push(p);
    int top = analyze_toclear.size();
    while (analyze_stack.size() > 0){
        assert(reason[var(analyze_stack.last())] != NULL);
        Clause& c = *reason[var(analyze_stack.last())]; analyze_stack.pop();

        for (int i = 1; i < c.size(); i++){
            Lit p  = c[i];
            if (!seen[var(p)] && level[var(p)] > 0){
                if (reason[var(p)] != NULL && (abstractLevel(var(p)) & abstract_levels) != 0){
                    seen[var(p)] = 1;
                    analyze_stack.push(p);
                    analyze_toclear.push(p);
                }else{
                    for (int j = top; j < analyze_toclear.size(); j++)
                        seen[var(analyze_toclear[j])] = 0;
                    analyze_toclear.shrink(analyze_toclear.size() - top);
                    return false;
                }
            }
        }
    }

    return true;
}


/*_________________________________________________________________________________________________
|
|  analyzeFinal : (p : Lit)  ->  [void]
|  
|  Description:
|    Specialized analysis procedure to express the final conflict in terms of assumptions.
|    Calculates the (possibly empty) set of assumptions that led to the assignment of 'p', and
|    stores the result in 'out_conflict'.
|________________________________________________________________________________________________@*/
void Solver::analyzeFinal(Lit p, vec<Lit>& out_conflict)
{
    out_conflict.clear();
    out_conflict.push(p);

    if (decisionLevel() == 0)
        return;

    seen[var(p)] = 1;

    for (int i = trail.size()-1; i >= trail_lim[0]; i--){
        Var x = var(trail[i]);
        if (seen[x]){
            if (reason[x] == NULL){
                assert(level[x] > 0);
                out_conflict.push(~trail[i]);
            }else{
                Clause& c = *reason[x];
                for (int j = 1; j < c.size(); j++)
                    if (level[var(c[j])] > 0)
                        seen[var(c[j])] = 1;
            }
            seen[x] = 0;
        }
    }

    seen[var(p)] = 0;
}


void Solver::uncheckedEnqueue(Lit p, Clause* from)
{
    assert(value(p) == l_Undef);
    assigns [var(p)] = toInt(lbool(!sign(p)));  // <<== abstract but not uttermost effecient
    level   [var(p)] = decisionLevel();
    reason  [var(p)] = from;
    trail.push(p);
}


/*_________________________________________________________________________________________________
|
|  propagate : [void]  ->  [Clause*]
|  
|  Description:
|    Propagates all enqueued facts. If a conflict arises, the conflicting clause is returned,
|    otherwise NULL.
|  
|    Post-conditions:
|      * the propagation queue is empty, even if there was a conflict.
|________________________________________________________________________________________________@*/

Clause* Solver::propagate()
{
    Clause* confl     = NULL;
    int     num_props = 0;

    while (qhead < trail.size()){
        Lit            p   = trail[qhead++];     // 'p' is enqueued fact to propagate.
        vec<Bclause>&  ws  = watches[toInt(p)];
        Bclause         *i, *j, *end;
        num_props++;
	Bclause B;
        
	for (i = j = ws, end = i + ws.size();  i != end;){
	  //B = *i++;
	  
	  if(value(i->a) == l_True){
	    *j++ = *i++;//B;
	  }
	  else{
	    B = *i++;
	    Clause & c = *B.cl;

            // Make sure the false literal is data[1]:
            Lit false_lit = ~p;
            if (c[0] == false_lit)
                c[0] = c[1], c[1] = false_lit;

            assert(c[1] == false_lit);

            // If 0th watch is true, then clause is already satisfied.
            Lit first = c[0];
            if (value(first) == l_True){
	      //printf("true\n");
	      //j->cl = 0;
	      B.a = first;
	      *j++ = B;
	      
	      //printf("\nj.cl=%d",(j->cl));
	    }else{
	      // Look for new watch:
	      for (int k = 2; k < c.size(); k++)
		if (value(c[k]) != l_False){
		  c[1] = c[k]; c[k] = false_lit;
		  if(value(c[1]) == l_True)
		    B.a = c[1];
		  
		  watches[toInt(~c[1])].push(B);
		  goto FoundWatch; }
	      
	      // Did not find watch -- clause is unit under assignment:
	      *j++ = B;
	      if (value(first) == l_False){
		confl = &c;
		qhead = trail.size();
		// Copy the remaining watches:
		while (i < end)
		  *j++ = *i++;
	      }else
		uncheckedEnqueue(first, &c);
            }
	  }
          FoundWatch:;
        }
	
        ws.shrink(i - j);
    }
    propagations += num_props;
    simpDB_props -= num_props;

    return confl;
}
/*_________________________________________________________________________________________________
|
|  reduceDB : ()  ->  [void]
|  
|  Description:
|    Remove half of the learnt clauses, minus the clauses locked by the current assignment. Locked
|    clauses are clauses that are reason to some assignment. Binary clauses are never removed.
|________________________________________________________________________________________________@*/
struct reduceDB_lt { bool operator () (Clause* x, Clause* y) { return x->size() > 2 && (y->size() == 2 || x->activity() < y->activity()); } };
void Solver::reduceDB(Cooperation* coop)
{
  int     i, j;
  double  extra_lim = cla_inc / learnts.size();    // Remove any clause below this activity
  
  sort(learnts, reduceDB_lt());
  for (i = j = 0; i < learnts.size() / 2; i++){
    if (learnts[i]->size() > 2 && !locked(*learnts[i]))
      removeClause(*learnts[i]);
    else
      learnts[j++] = learnts[i];
  }
  for (; i < learnts.size(); i++){
    if (learnts[i]->size() > 2 && !locked(*learnts[i]) && learnts[i]->activity() < extra_lim)
      removeClause(*learnts[i]);
    else
      learnts[j++] = learnts[i];
  }
  learnts.shrink(i - j);


  switch(coop->aimd){
    case 1:{
      Clauses_shared_Control(coop);
      nbConf = 0;
    }
    case 2:{
      Clauses_shared_Control_Q(coop);
      nbConf = 0;
    }
  }
  
}
void Solver::removeSatisfied(vec<Clause*>& cs)
{
    int i,j;
    for (i = j = 0; i < cs.size(); i++){
        if (satisfied(*cs[i]))
            removeClause(*cs[i]);
        else
            cs[j++] = cs[i];
    }
    cs.shrink(i - j);
}


/*_________________________________________________________________________________________________
|
|  simplify : [void]  ->  [bool]
|  
|  Description:
|    Simplify the clause database according to the current top-level assigment. Currently, the only
|    thing done here is the removal of satisfied clauses, but more things can be put here.
|________________________________________________________________________________________________@*/
bool Solver::simplify()
{

    assert(decisionLevel() == 0);

    if (!ok || propagate() != NULL)
        return ok = false;

    if (nAssigns() == simpDB_assigns || (simpDB_props > 0))
        return true;

    // Remove satisfied clauses:
    removeSatisfied(learnts);
    if (remove_satisfied)        // Can be turned off.
        removeSatisfied(clauses);

    // Remove fixed variables from the variable heap:
    order_heap.filter(VarFilter(*this));

    simpDB_assigns = nAssigns();
    simpDB_props   = clauses_literals + learnts_literals;   // (shouldn't depend on stats really, but it will do for now)

    return true;
}


/*_________________________________________________________________________________________________
|
|  search : (nof_conflicts : int) (nof_learnts : int) (params : const SearchParams&)  ->  [lbool]
|  
|  Description:
|    Search for a model the specified number of conflicts, keeping the number of learnt clauses
|    below the provided limit. NOTE! Use negative value for 'nof_conflicts' or 'nof_learnts' to
|    indicate infinity.
|  
|  Output:
|    'l_True' if a partial assigment that is consistent with respect to the clauseset is found. If
|    all variables are decision variables, this means that the clause set is satisfiable. 'l_False'
|    if the clause set is unsatisfiable. 'l_Undef' if the bound on number of conflicts is reached.
|________________________________________________________________________________________________@*/
lbool Solver::search(int nof_conflicts, int nof_learnts,Cooperation* coop)
{
    assert(ok);
    int         backtrack_level;
    int         conflictC = 0;
    vec<Lit>    learnt_clause;
    vec<Lit>    learnt_uipb;
    vec<Lit> resolution;
    starts++;
    Lit 	    conflict_lit = lit_Undef;
    bool first = true;
    int BT;
    nbExportedUnits = 0;
    nbExportedClauses = 0;
    
    for (;;){
      if(coop->End) {
		cancelUntil(0); 
		return l_Undef;
	  }//abort this thread, another thread has found out.
      
      //ASAP processing of external units clauses.
      if(decisionLevel() == 0)
		coop->importUnits(this);

      coop->importClauses(this);

      Clause* confl = propagate();
      if (confl != NULL){
	// CONFLICT
	conflicts++; conflictC++;
	if (decisionLevel() == 0) return l_False;
	
	first = false;
	
	learnt_clause.clear();
	if(threadId % 4 == 2)
	  Ranalyze(confl, learnt_clause, backtrack_level);
	else
	  analyze(confl, learnt_clause, backtrack_level);
	
	nbConf++;
	NewVarBump(learnt_clause[0]);

	if(threadId % 4 < 2 || (threadId % 4 == 3)){
	  b1+= decisionLevel();
	  b2+= (decisionLevel()-backtrack_level);
	  ++m1;
	}
	
	if(threadId % 4 == 3 || threadId % 4 == 0){
	  Lit q = trail[trail_lim[trail_lim.size()-1]];
	  assert(level[var(q)] == decisionLevel());
	  polarity[var(q)] = ~sign(q);//TODO
	}

	if (learnt_clause.size() == 1){// all threads
	  cancelUntil(backtrack_level);
	  assert(value(learnt_clause[0]) == l_Undef);
	  uncheckedEnqueue(learnt_clause[0]);
	  if(coop->limitExportClauses >= 1)
	    coop->exportUnit(this, learnt_clause[0]);
	  nb_unit++;
	}
	else{
	  if((threadId % 4 == 3 || threadId % 4 == 0) && searchReasonFromSuccess(learnt_uipb,BT, learnt_clause)){
	    //conflicts++; 
	    conflictC++;
	    
	    if(learnt_uipb.size() == 1){
	      cancelUntil(0);
	      uncheckedEnqueue(learnt_uipb[0]);
	      if(coop->limitExportClauses >= 1)
		coop->exportUnit(this, learnt_uipb[0]);
	    }
	    else{
	      cancelUntil(BT);
	      Clause* cs = Clause_new(learnt_uipb, true, threadId);
	      learnts.push(cs);
	      attachClause(*cs);
	      if(learnt_uipb.size() < learnt_clause.size())
		claBumpActivity(*cs);
	      
	      if(learnt_uipb.size() > learnt_clause.size()){
		Clause* c = Clause_new(learnt_clause, true, threadId);
			learnts.push(c);
			attachClause(*c);
			claBumpActivity(*cs);
	      }

	      uncheckedEnqueue(learnt_uipb[0], cs); 
	      //coop->exportClause(this, learnt_uipb);
	      
	    }// learnt_uipb.size() > 1
	  }//end threadId %4 == 0 && exists inverse-arcs
	  else{
	    cancelUntil(backtrack_level);
	    Clause* c = Clause_new(learnt_clause, true,threadId);
	    learnts.push(c);
	    attachClause(*c);
	    claBumpActivity(*c);
	    uncheckedEnqueue(learnt_clause[0], c);   
	  }//end threadId != 0 OR (threadId % 4 ==0 AND not exists inverse-arcs)

	  //all the threads!!!
	  coop->exportClause(this, learnt_clause);
	}
	varDecayActivity();
	claDecayActivity();
      }
    

    else{
            // NO CONFLICT
            if (nof_conflicts >= 0 && conflictC >= nof_conflicts){
	      
	      if(threadId %4 < 2 || threadId %4 == 3)
		{
		  n1 = (double)b1/m1;
		  n2 = (double)b2/m1;
		  
		  if(n1_old == 0) n1_old = n1;
		  if(somme == 0) somme = n2-0.1;

		  double kc = n2-somme;
		
		if((kc >= -0.7) && (kc <= 0.7))
		  tt++;
		
		if(tt == Res){
		  nb_tt++;
		  reduceDB(coop);
		  tt = 0;
		  Res = (int)(1.2*Res); 

		  for(int i = 0;i < nVars();i++)
		    activity[i] = 0;

		}
		if(n1 < n1_old){
		  C11 = 1500;
		  C22 = 1000;
		}else{
		  C11 = 1000;
		  C22 = 1500; //1500
		}
		
		if(somme == 0)
		  somme = 1;
		if(n2 == 0)
		  n2 = 1;
		if(threadId %4 == 0 || threadId %4 == 3){
		  if(n2 > somme)
		    restart_inc = C11/n2*cos(1-somme/n2);
		  else
		    restart_inc = C22/n2*cos(1-n2/somme); 
		}
		if(threadId %4 == 1){
		  if(n2 > somme)
		    restart_inc = C11/n2*cos(1+somme/n2);
		  else
		    restart_inc = C22/n2*cos(1+n2/somme);
		}
		
		
		if(threadId > 3)
		  restart_inc = restart_inc/2;
		
		somme = n2;
		n1_old = n1;
		b1 = 0; b2 = 0;m1 = 1;
		if(restart_inc < 0)
		  restart_inc = -restart_inc;
		if(restart_inc < 10) restart_inc = 10;

	      }//threadId %4 < 2

	      // Reached bound on number of conflicts:
              //progress_estimate = progressEstimate();
              cancelUntil(0);
              return l_Undef; 
	    }
	    
	    
	     if(nbConf >=  coop->periodic_Estimate_Length)
              {
		switch (coop->aimd){
		case 1:{
		  Clauses_shared_Control(coop);
		  nbConf = 0;}
		case 2:{
		  Clauses_shared_Control_Q(coop);
		  nbConf = 0;}
		}
              }



            // Simplify the set of problem clauses:
            if (decisionLevel() == 0 && !simplify())
                return l_False;

            if (nof_learnts >= 0 && learnts.size()-nAssigns() >= nof_learnts)
                // Reduce the set of learnt clauses:
	      reduceDB(coop);//todo

            Lit next = lit_Undef;
	    // New variable decision:
            decisions++;
            next = pickBranchLit(polarity_mode, random_var_freq, coop);
	    
            if (next == lit_Undef)
	      // Model found:
	      return l_True;
	    
            // Increase decision level and enqueue 'next'
            assert(value(next) == l_Undef);
            newDecisionLevel();
            uncheckedEnqueue(next);
         
        }
    }
}


double Solver::progressEstimate() const
{
    double  progress = 0;
    double  F = 1.0 / nVars();

    for (int i = 0; i <= decisionLevel(); i++){
        int beg = i == 0 ? 0 : trail_lim[i - 1];
        int end = i == decisionLevel() ? trail.size() : trail_lim[i];
        progress += pow(F, i) * (end - beg);
    }
    return progress / nVars();
}

/********  Luby *******************/
int get_luby(int i, int luby_unit)
{ //i restart
  if(i==1){return luby_unit;}  
  double k = log((double)i+1)/log(2.0);
  
  if(k==round(k)){
    return (int)(pow(2,k-1))*luby_unit;
  }else{
    k = (int)floor(k);
    return get_luby(i-(int)pow(2,k)+1, luby_unit);
  }
}


void Solver::printVarActivity(){
  printf("c \nThread%i, var activity=", threadId);
  for (int i = 0; i < nVars(); i++){
  }
}



void Solver::searchMostActivVar(Cooperation* C){


  int idRef = threadId % 4;

  for(int i = 0; i < C->solvers[idRef].nVars();i++){

    if(prior_lits.size() < prior_limit){
      prior_lits.push(Lit(i));
      activity[i] = C->solvers[idRef].activity[i];
    }
    
    else{
      for(int j = 0; j < prior_lits.size();j++){
	Lit q = prior_lits[j];
	
	if(C->solvers[idRef].activity[i] > activity[var(q)]){
	  prior_lits[j] = Lit(i);
	  activity[i] = C->solvers[idRef].activity[i];
	  break;
	}
      }
    }
  }

  for(int i = 0; i < prior_lits.size()-1;i++){
    Lit p = prior_lits[i];
    for(int j = i+1; j < prior_lits.size();j++){
      Lit q = prior_lits[j];
      if(activity[var(p)] < activity[var(q)]){
	prior_lits[i] = q;
	prior_lits[j] = p;
      }
    }
  }
}

void Solver::initFirstProGroup(){
  int ke = ((int)threadId/4 + 1);
  for(int i = nVars() / ke; i < nVars(); i++) 
    varBumpActivity(i);
  for(int i = 0; i< nVars();i++) 
    order_heap.update(i);
  for(int i = 0; i < nVars();i++) 
    activity[i] = 0;
}


extern char* Filename;
bool Solver::solve(const vec<Lit>& assumps, Cooperation* coop)
{
  receivedFrom  = new int[coop->nbThreads];
  keepedFrom  = new int[coop->nbThreads];
  for(int i = 0; i < coop->nbThreads;i++)
    {receivedFrom[i] = 0; keepedFrom[i] = 0;}
  
  n1_old = 0.0;
  nb_tt = 0;
  Res = 30;
  tt = 0;
  m1 = 0;
  b1 = 0;
  b2 = 0;
  n1 = 0.0;
  n2 = 0.0;
  nb_res = 0;
  somme = 0;
  nb_unit = 0;
  model.clear();
  conflict.clear();
  totNbExportedUnits = 0;
  totNbExportedClauses = 0;
  

  //Extension
  //initFirstProGroup();
  if (!ok) return false;


  assumps.copyTo(assumptions);
  
  double  nof_conflicts = restart_first;
  double  nof_learnts   = nClauses() * learntsize_factor;
  lbool   status        = l_Undef;
  
  //specialize restart_first
  switch(threadId % 4)
    {
    case 0:{// modulo scoring restart dynamic
      nof_conflicts = 3*restart_first;
      cla_inc = 1;
      polarity_mode = polarity_user;
      break;

    }
    
    case 1:{
      nof_conflicts = restart_first;
      cla_inc = 1;
      polarity_mode = polarity_user;
      break;
    }
    
    case 2:{
      nof_conflicts = 100;
      cla_inc = 1;
      polarity_mode = polarity_false;
      break;
    }
    
    case 3:{
      random_var_freq = 0.03;
      nof_conflicts = restart_first;
      cla_inc = 1;
      polarity_mode = polarity_user;
      break;
    }
   }
  
  coop->restarts = 0;
  int nrestarts = 1;
  // Search:
  while ((!coop->End)&&(status == l_Undef)){
    if (verbosity >= 1)
      printf("c | %3d | %6d | %9d | %7d %8d %8d | %8d %8d %6.0f | %6.3f %% | %6d |\n", 
	     coop->restarts, threadId, (int)conflicts, order_heap.size(), nClauses(), (int)clauses_literals, (int)nof_learnts, nLearnts(), (double)learnts_literals/nLearnts(), progress_estimate*100,
	     nbExportedUnits + nbExportedClauses), fflush(stdout);
    
    if(threadId > 3){
      prior_lits.clear();
      searchMostActivVar(coop);

      if(prior_limit < 20)
	prior_limit++;
      
    }
    
    status = search((int)nof_conflicts, (int)nof_learnts, coop);
    nrestarts++;
    coop->restarts++;
    
    totNbExportedUnits += nbExportedUnits;
    totNbExportedClauses += nbExportedClauses;
    
    //specialized restart policies
    switch(threadId % 4)
      {
      case 0:{
	nof_conflicts = restart_inc;
	nof_learnts   *= learntsize_inc;
	break;
      }
      case 1:{
	nof_conflicts =  restart_inc;
	nof_learnts   *= learntsize_inc;
	break;
      }
      case 2:{
	nof_conflicts *= restart_inc;
	nof_learnts   *= learntsize_inc;
	break;
      }
      case 3:{
	nof_conflicts = restart_inc;
	nof_learnts   *= learntsize_inc;
	break;
      }
      }
  }
 
    if (status == l_True){
        // Extend & copy model:
        model.growTo(nVars());
        for (int i = 0; i < nVars(); i++) model[i] = value(i);
        #ifndef NDEBUG
	//verifyModel(); 
	#endif
    }
	else if((status == l_False)){
		//assert(status == l_False);
		if (conflict.size() == 0)
		ok = false;
    }
    
    cancelUntil(0);
     

    return status == l_True;
}

    
//=================================================================================================
// Debug methods:


void Solver::verifyModel()
{
    bool failed = false;
    for (int i = 0; i < clauses.size(); i++){
        assert(clauses[i]->mark() == 0);
        Clause& c = *clauses[i];
        for (int j = 0; j < c.size(); j++)
            if (modelValue(c[j]) == l_True)
                goto next;

        printf("c unsatisfied clause: ");
        printClause(*clauses[i]);
        printf("\n");
        failed = true;
    next:;
    }

    assert(!failed);
    if(failed){ printf("BUG");exit(-1);}//todo
    printf("c Thread %i Verified %d original clauses.\n", threadId,clauses.size());
  
}


void Solver::checkLiteralCount()
{
    // Check that sizes are calculated correctly:
    int cnt = 0;
    for (int i = 0; i < clauses.size(); i++)
        if (clauses[i]->mark() == 0)
            cnt += clauses[i]->size();

    if ((int)clauses_literals != cnt){
        fprintf(stderr, "literal count: %d, real value = %d\n", (int)clauses_literals, cnt);
        assert((int)clauses_literals == cnt);
    }
}

bool Solver::searchReasonFromSuccess(vec<Lit>& out_learnt,int & out_btlevel, vec<Lit>& learned) {

  if(trail.size()==0)  return false;
  if(decisionLevel()==0) return false;

  Lit next = ~learned[1];
  
  bool top = true;
  for(int i = 2; i< learned.size();i++){
    if(level[var(learned[i])] == out_btlevel){
      top = false;
      break;
    }
  }
  if(!top) return false;
  
  for(int i = 0; i < Graph_Lit.size(); i++){
    seen2[var(Graph_Lit[i])] = 1;
    analyze_toclear.push(Graph_Lit[i]);
  }
  

  int nbseen = 0;
  
  if(next==lit_Undef) return false;
  vec<Lit> toto;
  
  for(int j = 0; j < learned.size();j++){
    seen3[var(learned[j])] = 1;
    toto.push(learned[j]);
  }
  bool ok_f = false;
  bool ok3 = false;bool okIci = false;
  
  vec<Bclause>&  wsa  = watches[toInt(~next)];
  
  for (int z = 0;z < wsa.size();z++){
    ok3 = false;okIci = false;
    Clause& ca = *(wsa[z]).cl ;
    assert(ca.size() > 1);
     if(ca.size()< 25){
       okIci = true;
     }
     if(okIci) {
       ok3 = true;
       nbseen = 0;
       for(int t = 0;t < ca.size();t++) {
	 if(var(ca[t]) != var(next)){

	   if(seen2[var(ca[t])]) nbseen++;
	   
	   if((value(ca[t])!=l_False)) {
	     ok3 = false; break;
	   }
	   if(!seen2[var(ca[t])]){
	     if(level[var(ca[t])] >= level[var(next)]){
	       ok3 = false; break;
	     }
	   }
	 }
       }
       
       
       if(nbseen == 0)
	 ok3 = false;
       
       if(ok3) {
	 
	 for(int i = 0; i < out_learnt.size();i++){
	   if(!seen3[var(out_learnt[i])]){
	     varBumpActivity(var(out_learnt[i]));
	     seen3[var(out_learnt[i])] = 1;
	     toto.push(out_learnt[i]);
	   }
	 }
	 bool ok5 = true;
	 out_learnt.clear();
	 vec<Lit> resolution;
	 seen[var(learned[0])] = 1;
	 analyze_toclear.push(learned[0]);
	 
	 for(int t = 0; t < learned.size(); t++){
	   if(t!=1){
	     out_learnt.push(learned[t]);
	     seen[var(learned[t])] = 1;
	     analyze_toclear.push(learned[t]);
	   }
	 }
	 
	 for(int i = 0; i < ca.size(); i++){
	   if(!seen[var(ca[i])]){
	     if((reason[var(ca[i])] != NULL) && (level[var(ca[i])] == decisionLevel())){
	       assert(seen2[var(ca[i])]);
	       resolution.push(ca[i]);
	     }
	     else{
	       if((level[var(ca[i])] > 0) && (level[var(ca[i])] < level[var(next)])){
		 out_learnt.push(ca[i]);
		 seen[var(ca[i])] = 1;
		 analyze_toclear.push(ca[i]);
	       }
	     }
	   }
	 }
	 if(resolution.size()== 0){
	   for(int j = 0; j < analyze_toclear.size(); j++)
	     seen[var(analyze_toclear[j])] = 0;
	   
	   int  lev = 0, r0 = 1;
	   Lit qb;
	   
	   for(int k = 1; k < out_learnt.size();k++){
	     if(level[var(out_learnt[k])] > lev){
	       lev = level[var(out_learnt[k])];
	       r0 = k;
	     }
	   }
	   
	   /*vec<Lit> toto;
	     
	   for(int j = 0; j < learned.size();j++){
	     seen[var(learned[j])] = 1;
	     toto.push(learned[j]);
	   }
	   
	   for(int i = 0; i < out_learnt.size();i++){
	     if(!seen[var(out_learnt[i])])
	       varBumpActivity(var(out_learnt[i]));
	   }
	 
	 for(int j = 0; j < toto.size();j++)
	   seen[var(toto[j])] = 0;
	   */
	   qb = out_learnt[1];
	   out_learnt[1] = out_learnt[r0];
	   out_learnt[r0] = qb;
	   out_btlevel = lev;
	   assert(lev < level[var(next)]);
	   ok_f = true;
	   break;
	 }
	 
	 if(resolution.size() > 0){
	   do {
	     Lit enCours = resolution.last();resolution.pop();
	     Clause &rEC = *(reason[var(enCours)]);
	     
	     for(int i = 1;i<rEC.size();i++) {
	       Lit aa = rEC[i];
	       if(!seen[var(aa)]) {
		 analyze_toclear.push(aa);
		 seen[var(aa)] = 1;
		 if( !seen2[var(aa)] || (reason[var(aa)]==NULL) || (level[var(aa)]!= decisionLevel())) {
		   
		   if(level[var(aa)] > 0)
		     out_learnt.push(aa);
		   if((level[var(aa)] ==level[var(next)]) ){
		     ok5 = false;
		   }
		 } else {
		   resolution.push(aa);
		 }
	       }
	     }
	   } while(resolution.size() > 0);
	   
	   for(int j = 0; j < analyze_toclear.size(); j++)
	     seen[var(analyze_toclear[j])] = 0;
	   
	   if(ok5){
	     int lev = 0;
	     int u = 1;
	     Lit pp;
	     
	     for(int t = 1; t < out_learnt.size(); t++){
	       if(level[var(out_learnt[t])] > lev){
		 lev =  level[var(out_learnt[t])];
		 u = t;
	       }
	     }
	     pp = out_learnt[1];
	     out_learnt[1] = out_learnt[u];
	     out_learnt[u] = pp;
	     
	     if((out_learnt.size() < (learned.size()+30)) && (lev < level[var(next)])){
	       
	       /* vec<Lit> toto;
	       
	       for(int j = 0; j < learned.size();j++){
		 seen[var(learned[j])] = 1;
		 toto.push(learned[j]);
	       }
	       
	       for(int i = 0; i < out_learnt.size();i++){
		 if(!seen[var(out_learnt[i])])
		   varBumpActivity(var(out_learnt[i]));
	       }
	       
	       for(int j = 0; j < toto.size();j++)
		 seen[var(toto[j])] = 0;
	       */
	       out_btlevel = lev;
	       ok_f = true;
	       break;
	     }
	   }
	 }
	 
	 for(int j = 0; j < analyze_toclear.size(); j++)
	   seen[var(analyze_toclear[j])] = 0;
       }
     }
  }
  for(int j = 0; j < toto.size();j++)
    seen3[var(toto[j])] = 0;
  
  for(int j = 0; j < analyze_toclear.size(); j++){
    seen2[var(analyze_toclear[j])] = 0;
    seen[var(analyze_toclear[j])] = 0;
  }
  analyze_toclear.clear();
  return ok_f;  
}

/////////////////////////////////////////////////////////////////////////////////////////
	
void Solver::NewVarBump(Lit assert_lit){
  vec<Lit> resolution;
if(decisionLevel() > 1 && reason[var(conflict_lit)] != NULL && 
	   (conflict_lit != lit_Undef) && (reason[var(assert_lit)] != NULL)){
	  
	  seen[var(conflict_lit)] = 1;
	  analyze_toclear.push(conflict_lit);
	  resolution.clear();
	  
	  resolution.push(conflict_lit);
	  varBumpActivity(var(conflict_lit));
	  
	  do {
	    Lit enCours = resolution.last();resolution.pop();
	    Clause &rEC = *(reason[var(enCours)]);
	    
	    for(int i = 1;i< rEC.size();i++) {
	      Lit aa = rEC[i];
	      if(!seen[var(aa)]) {
		analyze_toclear.push(aa);
		seen[var(aa)] = 1;
		if((reason[var(aa)]!=NULL) && (level[var(aa)] == decisionLevel())) {
		  varBumpActivity(var(aa));
		  resolution.push(aa);
		}
	      }
	    }
	  } while(resolution.size() > 0);
	  
	  
	  for(int i = 0; i < analyze_toclear.size();i++){
	    seen[var(analyze_toclear[i])] = 0;
	  }
	  analyze_toclear.clear();
	}
}

/////////////////////////////////////////////////////
//		Multi-threading Search
/////////////////////////////////////////////////////
#define MAX_EXTRA_CLAUSES 2000
#define MAX_EXTRA_UNITS 1000
#define AIMD_A 0.125
#define AIMD_B 8
#define THROUGHPUT 0.5
#define PERIOD 10000
Cooperation::Cooperation(int n, int l, int d){
	nbThreads = n;
	limitExportClauses = l;
	solvers = new Solver [nbThreads];


        //Control-based clause sharing  
	e                        = new double *[nbThreads];
	quality                  = new double *[nbThreads];
        average_e                = new double *[nbThreads];
        nb_compute_e             = new double [nbThreads];
        a                        = AIMD_A;
        b                        = AIMD_B;
        B                        = THROUGHPUT;
        periodic_Estimate_Length = PERIOD;

	aimd                     = d;

	for(int t = 0; t < nbThreads; t++){
            e[t]                      = new double[nbThreads];
            quality[t]                = new double[nbThreads];
            average_e[t]              = new double[nbThreads];

            for(int i = 0; i < nbThreads; i++){
	      nb_compute_e[t]        = 0;
	      average_e[t][i]        = 0;
	      e[t][i]                = l;
	      quality[t][i]          = 0;
            }
        }

	//units, extraUnits[send][recv][nbExtraUnits[send][recv]] = unit
	extraUnits = new Lit** [nbThreads];
	headExtraUnits = new int* [nbThreads];
	tailExtraUnits = new int* [nbThreads];
	
	for(int t = 0; t < nbThreads; t++){
	  extraUnits[t] = new Lit* [nbThreads];
	  headExtraUnits[t] = new int [nbThreads];
	  tailExtraUnits[t] = new int [nbThreads];
	  //extraUnits[t] = 0;
	  
	  for(int k = 0; k < nbThreads; k++){
	    extraUnits[t][k] = new Lit [MAX_EXTRA_UNITS];
	    headExtraUnits[t][k] = 0;
	    tailExtraUnits[t][k] = 0;
	  }
	}

	//clauses, extraClauses[send][recv][nbExtraClauses[send][recv]] = clause
	extraClauses = new Lit*** [nbThreads];
	headExtraClauses = new int* [nbThreads];
	tailExtraClauses = new int* [nbThreads];

	for(int t = 0; t < nbThreads; t++){
	  extraClauses[t] = new Lit** [nbThreads];
	  headExtraClauses[t] = new int [nbThreads];
	  tailExtraClauses[t] = new int [nbThreads];
 
	  for(int k = 0; k < nbThreads; k++){
	    extraClauses[t][k] = new Lit* [MAX_EXTRA_CLAUSES];
	    headExtraClauses[t][k] = 0;
	    tailExtraClauses[t][k] = 0;
	  }
	}

	End = false;
}


void Cooperation::exportUnit(Solver* s, Lit unit){
  int id = s->threadId;
  for(int t = 0; t < nbThreads; t++)
    if(t != id){
      int ind = tailExtraUnits[id][t];
      if(((ind+1)%MAX_EXTRA_UNITS) == headExtraUnits[id][t]){
	//printf("\n%d FULL", id);
	break;
      }
      extraUnits[id][t][ind++].x = unit.x;
      //printf("\n%d EXPORT unit.", id);

      if(ind == MAX_EXTRA_UNITS)
	ind = 0;

      tailExtraUnits[id][t] = ind;
    }
  s->nbExportedUnits++;
}

void Cooperation::importUnits(Solver* s){ 
  int id = s->threadId;
  for(int t = 0; t < nbThreads; t++)
   if(t != id){
     int head = headExtraUnits[t][id];
     int tail = tailExtraUnits[t][id];
     
     if(head == tail) break;

     int end = tail; 

     if(tail < head) end = MAX_EXTRA_UNITS;

     for(int i = head; i < end; i++)
       if(s->value(extraUnits[t][id][i]) == l_Undef)
	 s->uncheckedEnqueue(extraUnits[t][id][i]);
     
     if(tail < head)
       for(int i = 0; i < tail; i++)
	 if(s->value(extraUnits[t][id][i]) == l_Undef)
	   s->uncheckedEnqueue(extraUnits[t][id][i]);

     head = tail;
     if(head == MAX_EXTRA_UNITS) head = 0;
     headExtraUnits[t][id] = head;
   }
}


void Cooperation::exportClause(Solver* s, vec<Lit>& learnt){

  int id = s->threadId;

  for(int t = 0; t < nbThreads; t++)
    
    if(t != id){
      
      
      if( ((id <= 3) && ((t < 4)  || (t % 4 == id))) || ((id >= 3) && (id % 4 == t))){
	
	if(learnt.size() <= e[id][t]){
	  int ind = tailExtraClauses[id][t];
	  if(((ind+1)%MAX_EXTRA_CLAUSES) == headExtraClauses[id][t]){
	    break;
	  }
	  
	  extraClauses[id][t][ind] = new Lit [learnt.size() + 1];
	  extraClauses[id][t][ind][0] = Lit(learnt.size());
	  
	  for(int j = 0; j < learnt.size(); j++)
	    extraClauses[id][t][ind][j + 1] = learnt[j];
	  
	  ind++;
	  
	  if(ind == MAX_EXTRA_CLAUSES) ind = 0;
	  
	  tailExtraClauses[id][t] = ind;
	}
	}	
    }

  s->nbExportedClauses++; 
}		      
	
void Cooperation::addC(Solver* s, Lit* lt, int sender){
  vec<Lit> l;
  int size = var(lt[0]);
  int nbactiv=0;
  for(int j = 1; j < size+1; j++){
     l.push(lt[j]);
     if((aimd >= 2) && (s->activity[var(lt[j])] > s->max_act/2))
       nbactiv++;
  }
  Clause* c2 = Clause_new(l, true, s->threadId);
  s->learnts.push(c2);
  s->attachClause(*c2);
  s->claBumpActivity(*c2);
  

  switch (aimd){
  case 1:{
    s->receivedFrom[sender]++;
  }
  case 2:{
    s->receivedFrom[sender]++;
    if(nbactiv >= (double)size/3)
      s->keepedFrom[sender]++;
  }
  }

  
  delete [] lt;
  
}		      


void Cooperation::importClauses(Solver* s){
  int id = s->threadId;
  for(int t = 0; t < nbThreads; t++)
   if(t != id){
     int head = headExtraClauses[t][id];
     int tail = tailExtraClauses[t][id];
     
     if(head == tail) break; 

     int end = tail; 

     if(tail < head) end = MAX_EXTRA_CLAUSES;

     for(int i = head; i < end; i++)
       addC(s, extraClauses[t][id][i], t);
     
     if(tail < head)
      for(int i = 0; i < tail; i++)
	addC(s, extraClauses[t][id][i], t);
     
     head = tail;
     if(head == MAX_EXTRA_CLAUSES) head = 0;
     headExtraClauses[t][id] = head;
   }
}

void Solver::Clauses_shared_Control(Cooperation * coop){

  coop->nb_compute_e[threadId]++;
  
  int receivedFromAll = 0;
  for(int i = 0; i < coop->nbThreads;i++)
    receivedFromAll += receivedFrom[i];
  
  if((double)receivedFromAll > coop->B * nbConf)
    {
      
      for(int i = 0;i < coop->nbThreads;i++)
        if(i != threadId)
	  {
	    coop->quality[i][threadId]          = (double) 100* (keepedFrom[i]+1)/(receivedFrom[i]+1) ;
	    //double              Q               = coop->quality[i][threadId];
	    coop->e[i][threadId]               -= coop->a * coop->e[i][threadId];
	    coop->average_e[i][threadId]       +=  coop->e[i][threadId];
	  }
    }
  else
    {
      for(int i = 0;i < coop->nbThreads;i++)
        if(i != threadId)
	  {
	    coop->quality[i][threadId]          = (double) 100* (keepedFrom[i]+1)/(receivedFrom[i]+1) ;
	    //double              Q               = coop->quality[i][threadId];
	    coop->e[i][threadId]              +=   coop->b / coop->e[i][threadId];
	    coop->average_e[i][threadId]      +=  coop->e[i][threadId];
	  }
    }

  if(verbosity >= 1 && threadId == 0){
    for(int i = 0; i < coop->nbThreads;i++)
      if(i != threadId)
        printf("\t e%d%d = [%2.2f] (%2.2f %4d -- %4d)", i,threadId, coop->e[i][threadId], coop->quality[i][threadId], keepedFrom[i], receivedFrom[i]);
    printf("\n");
  }

  for(int i = 0; i < coop->nbThreads;i++)
    {receivedFrom[i] = 0;}
  
}



void Solver::Clauses_shared_Control_Q(Cooperation * coop){

  coop->nb_compute_e[threadId]++;
  
  int receivedFromAll = 0;
  for(int i = 0; i < coop->nbThreads;i++)
    if(i != threadId)
      receivedFromAll += receivedFrom[i];

  if((double)receivedFromAll > coop->B * nbConf)
    {
      
      for(int i = 0;i < coop->nbThreads;i++)
        if(i != threadId){
	  coop->quality[i][threadId]          = (double) 100* (keepedFrom[i]+1)/(receivedFrom[i]+1) ;
	  double              Q               = coop->quality[i][threadId];
	  coop->e[i][threadId]               -=   coop->a *(1 - Q/100) * coop->e[i][threadId];
	  coop->average_e[i][threadId]      +=  coop->e[i][threadId];
	}
    }
  else
    {
      for(int i = 0;i < coop->nbThreads;i++)
        if(i != threadId){
	  coop->quality[i][threadId]          = (double) 100* (keepedFrom[i]+1)/(receivedFrom[i]+1) ;
	  double              Q               = coop->quality[i][threadId];
	  coop->e[i][threadId]              +=   (Q/100)*coop->b / coop->e[i][threadId];
	  coop->average_e[i][threadId]      +=  coop->e[i][threadId];
	}
    }
  
  if(verbosity >= 1 && threadId == 0  && coop->nbThreads > 1){
    for(int i = 0; i < coop->nbThreads;i++)
      if(i != threadId)
        printf("c e%d%d = [%2.2f] (%2.2f %4d -- %4d)", i,threadId, coop->e[i][threadId], coop->quality[i][threadId], keepedFrom[i], receivedFrom[i]);
    printf("\n");
  }
  
  for(int i = 0; i < coop->nbThreads;i++)
    {receivedFrom[i] = 0;keepedFrom[i] = 0;}
}



void Cooperation::print(){
  int sumExportUnits = 0;
  int sumExportClauses = 0;
 
 for(int t = 0; t < nbThreads; t++){
    sumExportUnits += solvers[t].totNbExportedUnits;
    sumExportClauses += solvers[t].totNbExportedClauses;
  }
  printf("\nc #exported units = %d, #exported clauses (<= %d) = %d", sumExportUnits, limitExportClauses, 
	 sumExportClauses);


  double sum = 0;
  int nb     = 0;
  for(int i = 0; i < nbThreads;i++)
    {
      for(int j = 0; j < nbThreads;j++)
        if(i != j){
          //  printf("m%d%d=%2.2f    ",j,i,average_e[j][i]/nb_compute_e[i]); 
          nb++;
          sum += average_e[j][i]/nb_compute_e[i];
	  //if(nb_compute_e[i] == 0){exit(0);}
        }
      //printf("\n");
    }
  if(nb == 0) nb = 1;

  if(aimd)
    printf("\nc aimd = %d (a = %g, b = %g, period = %i, throughput = %g), e-bar = %f", aimd, a, b, periodic_Estimate_Length, B, sum/nb);


}
